<?php
declare(strict_types=1);

namespace Jiwei\EasyHttpSdk;

use GuzzleHttp\Client;
use GuzzleHttp\Exception\GuzzleException;
use GuzzleHttp\Exception\TransferException;
use GuzzleHttp\HandlerStack;
use GuzzleHttp\Psr7\Request;
use InvalidArgumentException;
use Jiwei\EasyHttpSdk\Exception\ApplicationException;
use Jiwei\EasyHttpSdk\Exception\SdkException;
use Jiwei\EasyHttpSdk\Exception\TimeOutExcetpion;
use Jiwei\EasyHttpSdk\Http\SdkRequest;
use Jiwei\EasyHttpSdk\Middleware\JwtMiddleware;
use Psr\Cache\CacheItemPoolInterface;
use Psr\Http\Client\ClientInterface;
use Psr\Http\Message\RequestInterface;
use Psr\Http\Message\ResponseInterface;
use Psr\Log\LoggerInterface;
use Psr\Cache\InvalidArgumentException as CacheInvalidArgumentException;
use RuntimeException;
use Symfony\Component\Cache\Adapter\FilesystemAdapter;


class Application implements ClientInterface
{

    private const TOKEN_CACHE_KEY = "Auth.%s";

    private const EXPIRES_AT = 30000;

    /** @var Option $option */
    private $option;

    /** @var string jwtToken */
    private $jwtToken = "";

    /** @var null|LoggerInterface 日志记录器 */
    protected $logger = null;

    /** @var null|Client Guzzle客户端 */
    protected $client = null;

    /** @var null|CacheItemPoolInterface PS-6缓存 */
    protected $cache = null;

    /** @var array<string, mixed> 最后的一次请求调用上下文 */
    protected $lastRequestContext = [];

    /**
     * 初始化一个 SDK 应用，目前 config 中必填的有 app_key 和 app_secret, 剩余 stage 默认为 development, debug 默认为 false, timeout 默认为3.0,
     * 最后也可以传入 PSR 规范下的日志接口, 当然也可以不进行日志记录,
     * Cache的需要实现PSR规范的缓存接口, 默认使用 symfony/cache 的实现, 也可以自行替换为 Redis (需要实现 PSR-6 ),
     *
     * @param Option $option
     * @param LoggerInterface|null $logger
     * @param CacheItemPoolInterface|null $cache
     */
    public function __construct(
        Option                  $option,
        ?LoggerInterface        $logger = null,
        ?CacheItemPoolInterface $cache = null
    )
    {

        if (empty($option->getAppId())) {
            throw new RuntimeException("请完善配置信息中的App Key");
        }

        if (empty($option->getAppSecret())) {
            throw new RuntimeException("请完善配置信息中的App Secret");
        }

        $this->option = $option;

        $base_url = $option->getBaseUrl();

        if (empty($base_url)) {
            throw new RuntimeException("请检查配置信息中的Base Url");
        }

        if ($logger) {
            $this->logger = $logger;
        }

        $this->cache = new FilesystemAdapter();
        if ($cache) {
            $this->cache = $cache;
        }

        $this->client = new Client([
            'headers' => [
                "Content-Type" => 'application/json; charset=utf8',
            ],
            "base_uri" => $base_url,
            "timeout" => $this->option->getTimeOut(),
            "http_errors" => false,
            "debug" => $this->option->isDebug(),
        ]);
    }

    /**
     * 发送SDK的请求
     *
     * @param RequestInterface $request
     * @return ResponseInterface
     * @throws ApplicationException
     * @throws GuzzleException
     *
     */
    public function sendRequest(RequestInterface $request): ResponseInterface
    {

        $stack = HandlerStack::create();

        $this->lastRequestContext = [
            'request' => $request,
            "stage" => $this->option->getStage(),
        ];
        $start = microtime(true);
        // 应该使用的是 SDKRequest
        if ($request instanceof SdkRequest) {

            $this->lastRequestContext['request_id'] = $request->getUuid();
            $this->lastRequestContext['args'] = $request->getArgs();
            $this->lastRequestContext['endpoint'] = $request->getEndpoint();
            $this->lastRequestContext['action'] = $request->getAction();

            $start = $request->getStartAt();
            $middlewares = $request->getMiddlewares();
            foreach ($middlewares as $middleware) {
                $stack->push($middleware);
            }
            if ($request->authorization()) {
                if ($this->jwtToken === "") {
                    try {
                        $this->jwtToken = $this->getAccessTokenFromCache($this->option->getAppId(), $this->option->getAppSecret());
                    } catch (RuntimeException $exception) {
                        if ($this->logger) {
                            $this->logger->error("sdk error, bad Auth with: " . $exception->getMessage());
                        }
                        throw new TimeOutExcetpion($request);
                    } catch (InvalidArgumentException $exception) {
                        if ($this->logger) {
                            $this->logger->error("sdk error, bad Auth with:" . $exception->getMessage());
                        }
                        throw new ApplicationException($exception->getMessage(), $request, [], $exception);
                    }
                }
                $stack->push(new JwtMiddleware($this->jwtToken));
            }
        }

        $options = [
            'handler' => $stack
        ];

        try {
            $response = $this->client->send($request, $options);
        } catch (TransferException $exception) {
            $this->lastRequestContext['expend'] = sprintf("%dms", microtime(true) * 1000 - $start * 1000);
            if ($this->logger) {
                $this->logger->error("time out!");
            }
            if (!($request instanceof SdkRequest)) {
                throw $exception;
            }
            throw new TimeOutExcetpion($request, $exception);
        } catch (GuzzleException $exception) {
            if ($this->logger) {
                $this->logger->error("sdk etrror, bad client!");
            }
            if (!($request instanceof SdkRequest)) {
                throw $exception;
            }
            throw new ApplicationException($exception->getMessage(), $request, [], $exception);
        }

        $this->lastRequestContext['expend'] = sprintf("%dms", microtime(true) * 1000 - $start * 1000);
        $this->lastRequestContext['http_status'] = $response->getStatusCode();
        $this->lastRequestContext['response'] = $response;
        return $response;
    }

    /**
     * 解析SDK的响应结果，并处理异常
     *
     * @param SdkRequest $request
     * @param ResponseInterface $response
     * @return array<string, mixed>
     * @throws ApplicationException
     * @throws SdkException
     */
    public function resultDecode(SdkRequest $request, ResponseInterface $response): array
    {

        $context = $this->lastRequestContext;

        if ($context['request']) {
            unset($context['request']);
        }
        if ($context['response']) {
            unset($context['response']);
        }

        if ($response->getStatusCode() >= 500) {
            if ($this->logger) {
                $this->logger->error("sdk error, paypal service fail", $context);
            }
            throw new SdkException("SDK Upstram Failed", $request, $response, []);
        }

        if (in_array($response->getStatusCode(), [404, 405])) {
            if ($this->logger) {
                $this->logger->error("sdk error, endpoint not found", $context);
            }
            throw new ApplicationException("Api Endpoint Not Found.", $request, []);
        }

        $rpcResult = [];

        if ($response->getStatusCode() != 304) {
            $responseInfo = $response->getBody()->getContents();
            $rpcResult = json_decode($responseInfo, true);
            if (json_last_error()) {
                if ($this->logger) {
                    $this->logger->error("sdk error, bad response content.", $context);
                }
                throw new ApplicationException("Content Format error.", $request, []);
            }
        }

        $errorName = $rpcResult['name'] ?? "";
        // Paypal的API错误结果是通用的，所以这里简单处理一下
        if (!empty($errorName) && $response->getStatusCode() >= 400) {
            $errorMessage = sprintf("%s : %s", $errorName, $rpcResult['message'] ?? "");
            $this->lastRequestContext['error'] = $errorMessage;
            $this->lastRequestContext['error_detail'] = $rpcResult;
            if ($this->logger) {
                $this->logger->warning("Error", $this->lastRequestContext);
            }
            throw new SdkException($errorMessage, $request, $response, $rpcResult);
        }

        $version = $response->getHeader('ETag');

        if (!empty($version)) {
            $rpcResult = array_merge($rpcResult, [
                'version' => $version[0]
            ]);
        }
        if ($this->logger) {
            $this->logger->info("Success", $context);
        }


        return $rpcResult;
    }

    /**
     * 获取上一次请求的上下文
     *
     * @return array<string, mixed>
     */
    public function getLastRequestContext(): array
    {
        return $this->lastRequestContext;
    }

    /**
     * 从缓存中获取accessToken 使用PSR-6缓存接口
     *
     * @param string $app_key
     * @param string $app_secret
     * @return string
     */
    private function getAccessTokenFromCache(string $app_key, string $app_secret): string
    {
        $cacheKey = sprintf(self::TOKEN_CACHE_KEY, $app_key);
        try {
            $currentToken = $this->cache->getItem($cacheKey);
        } catch (CacheInvalidArgumentException $e) {
            throw new InvalidArgumentException("cache use invalid args" . $e->getMessage());
        }

        if (!$currentToken->isHit()) {
            $token = $this->getAccessToken($app_key, $app_secret);

            $currentToken->set($token)->expiresAfter(self::EXPIRES_AT);
            $this->cache->save($currentToken);
        }

        return $currentToken->get();
    }

    /**
     * 鉴权
     *
     * @param string $app_key
     * @param string $app_secret
     * @return string
     */
    private function getAccessToken(string $app_key, string $app_secret): string
    {
        try {
            $request = $this->option->getAuthMoudel()($app_key, $app_secret);
            $response = $this->client->send($request);
        } catch (TransferException $exception) {
            throw new RuntimeException('time out!');
        } catch (GuzzleException $exception) {
            if ($this->logger) {
                $this->logger->error("sdk error, Auth error");
            }
            throw new InvalidArgumentException("auth error." . $exception->getMessage());
        }

        if ($response->getStatusCode() >= 400) {
            throw new InvalidArgumentException("invalid account.");
        }

        $rpcResult = json_decode($response->getBody()->getContents(), true);

        if (json_last_error()) {
            throw new InvalidArgumentException("invalid auth response");
        }

        $token = $rpcResult["access_token"] ?? "";

        if ($token === "") {
            throw new InvalidArgumentException("invalid account.");
        }
        return $token;
    }
}
